import io
import os
from typing import Optional

import pandas as pd
import sqlalchemy
from azure.storage.blob import BlobClient, BlobServiceClient

from ..common.jragbeer_common_data_eng import dagster_logger, data_path


def adls_upload_folder(
    directory: str = data_path, file_search_string: Optional[str] = None
) -> None:
    """
    This function uploads each file in the specified directory to ADLS.
    :param directory: the directory with the files
    :param file_search_string: a string used to filter the files in the directory
    :return: None
    """
    if file_search_string:
        file_list = [i for i in os.listdir(directory) if file_search_string in i]
    else:
        file_list = [i for i in os.listdir(directory)]

    for each_file in file_list:
        adls_upload_file(directory + each_file, each_file)


def adls_upload_file(file_path: str, blob_name: str = "blob_path_or_name") -> None:
    """
    This function uploads a file to ADLS.
    :param file_path: The path of the file.
    :param blob_name: The name of the file in ADLS
    :return: None
    """
    with open(file_path, "rb") as file_to_upload:
        blob_service_client = BlobServiceClient.from_connection_string(
            os.getenv("adls_connection_string")
        )
        # Instantiate a new ContainerClient
        container_client = blob_service_client.get_container_client(os.getenv("adls_container_name"))
        # Instantiate a new BlobClient
        blob_client = container_client.get_blob_client(blob_name)
        # upload data
        blob_client.upload_blob(file_to_upload, blob_type="BlockBlob", overwrite=True)
        dagster_logger.info(f"""{blob_name} is uploaded to BLOB storage (container: {os.getenv("adls_container_name")})""")


def adls_upload_sql_table(
    engine: sqlalchemy.engine.base.Engine,
    table_name: str = "name_of_sql_table",
) -> None:
    """
    This function reads a table in the database and then outputs it as a file in ADLS.
    :param engine: the SQL Alchemy engine
    :param table_name: the name of the table in SQL
    :return:
    """
    df = pd.read_sql(f"select * from {table_name} ", engine)
    output = df.to_parquet()
    blob = BlobClient.from_connection_string(
        os.getenv("adls_connection_string"), container_name=os.getenv("adls_container_name"), blob_name=table_name
    )
    blob.upload_blob(
        output,
        overwrite=True,
    )
    dagster_logger.info(
        f"""Latest {table_name} is uploaded to BLOB storage (container: {os.getenv("adls_container_name")})"""
    )


def adls_upload_df(
    idf_: pd.DataFrame,
    blob_name: str = "name_of_blob",
    container_name: str = os.getenv("adls_container_name", "adls_container_name"),
) -> None:
    """
    This function takes a dataframe and converts all columns to ADLS friendly data types and then uploads it to ADLS.
    :param df: the dataframe to be uploaded
    :param blob_name: the name of the output file in ADLS
    :return: None
    """
    for col_type, col_name in zip(idf_.dtypes, idf_.columns):
        if col_type == "object":
            idf_[col_name] = idf_[col_name].astype(str)
            idf_[col_name] = [i[:4000] for i in idf_[col_name]]
    output = (
        idf_.reset_index(drop=True)
        .drop_duplicates()
        .replace({"nan": None})
        .to_parquet(index=False)
    )
    blob = BlobClient.from_connection_string(
        os.getenv("adls_connection_string"),
        container_name=str(container_name).lower(),
        blob_name=blob_name,
    )
    blob.upload_blob(
        output,
        overwrite=True,
        timeout=14400,
        connection_timeout=14400,
    )
    dagster_logger.info(
        f"Latest {blob_name} is uploaded to BLOB storage (container: {container_name})"
    )

def adls_download_parquet_file(blob_name: str, input_container_name : str = os.getenv("adls_container_name", "adls_container_name")) -> pd.DataFrame:
    """
    This function downloads a parquet file from ADLS.
    @param blob_name: the name of the blob inside of ADLS
    @return: the dataframe that's downloaded
    """
    blob = BlobClient.from_connection_string(
        os.getenv("adls_connection_string"), container_name=input_container_name, blob_name=blob_name
    )
    # if there's a file that exists in the BLOB
    if blob.exists():
        # Download blob as StorageStreamDownloader object (stored in memory)
        downloaded_blob = blob.download_blob()
        # check if we can read the parquet file, if we can't probably not a parquet file
        try:
            bytes_io = io.BytesIO(downloaded_blob.readall())
            idf = pd.read_parquet(bytes_io)
            return idf
        except Exception:
            dagster_logger.info("Reading BLOB did not work.")
def adls_download_text_file(blob_name: str, input_container_name : str = os.getenv("adls_container_name", "adls_container_name"), save_file:Optional[str]=None) -> Optional[str]:
    """
    This function downloads a parquet file from ADLS.
    @param blob_name: the name of the blob inside of ADLS
    @return: str of the text file
    """
    blob = BlobClient.from_connection_string(
        os.getenv("adls_connection_string", "adls_connection_string"), container_name=input_container_name, blob_name=blob_name
    )
    # if there's a file that exists in the BLOB
    text_data = None
    if blob.exists():
        # Download blob as StorageStreamDownloader object (stored in memory)
        downloaded_blob = blob.download_blob()
        # check if we can read the parquet file, if we can't probably not a parquet file
        try:
            text_data = downloaded_blob.readall().decode("utf-8")
            if save_file:
                with open(save_file, 'w') as f:
                    f.write(text_data)
            return text_data
        except Exception:
            dagster_logger.info("Reading BLOB did not work.")
            return text_data
    return text_data
def adls_replace_nans_with_nulls_in_dfs() -> None:
    """
    Get all of the blobs in the main container and convert the 'nan's in them to None. Then reupload the blob.
    @return: None
    """
    blob_svc = BlobServiceClient.from_connection_string(conn_str=os.getenv("adls_connection_string"))
    dagster_logger.info("\nList blobs in the container")
    container_client = blob_svc.get_container_client(os.getenv("adls_container_name"))
    blob_list = container_client.list_blobs()
    for blob in blob_list:
        if blob.name.endswith(".parquet"):
            try:
                dagster_logger.info(blob.name)
                pdf = adls_download_parquet_file(blob.name)
                adls_upload_df(pdf, blob.name)
            except Exception as ee:
                dagster_logger.info(ee)



import os
import time
from pprint import pprint
from typing import Any, Callable, Literal, Optional

import dask
import pandas as pd
import paramiko  # type: ignore
import sqlalchemy
from dask.distributed import get_client

from ..common.jragbeer_common_data_eng import dagster_logger, error_handling
from ..ubuntu.jragbeer_common_ubuntu import (create_yaml_from_dict,
                                             execute_cmd_ubuntu_sudo,
                                             execute_script_with_cmd,
                                             get_remote_process_ids_ubuntu,
                                             give_write_permission_to_folder,
                                             kill_remote_ubuntu_process_ids)

CLUSTER_TYPE = Literal["SERIAL", "LOCAL", "DISTRIBUTED"]

def update_dask_environment_vars_local(env_dict):
    folder = "/etc/dask"
    execute_cmd_ubuntu_sudo(f"mkdir {folder}")
    give_write_permission_to_folder(folder)
    create_yaml_from_dict(env_dict, "/etc/dask/dask.yaml")

def create_dask_scheduler(hostname:str, username:str, password:str,
                          port:int=8786, dashboard_port:int=8787) -> tuple[str,str]:
    # SSH into the remote machine
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostname, username=username, password=password)

    command = (
        "docker exec finance_scheduler bash -c "
        "'cd /opt/dagster/app && "
        "source .venv/bin/activate && "
        f"nohup dask scheduler --host 0.0.0.0 --port {port} "
        f"--dashboard-address :{dashboard_port} > /tmp/dask_scheduler.log 2>&1 &'"
    )

    stdin, stdout, stderr = ssh.exec_command(command)
    dagster_logger.info(f"stdin:{stdin} | stdout:{stdout} | stderr:{stderr}")
    dagster_logger.info(f"Dask Scheduler created on remote machine : {hostname}:{port}")
    time.sleep(1)
    # Fetch the scheduler log (not worker log)
    _, log_out, log_err = ssh.exec_command("tail -n 75 /tmp/dask_scheduler.log")
    print(log_out.read().decode(), log_err.read().decode())
    return log_out, log_err

def create_dask_worker_on_scheduler(hostname:str, username:str, password:str,
                                     scheduler_ip:str='localhost', scheduler_port:int=8786,
                                     nworkers: int=1, mem_limit: str = '1GB',
                                     worker_port_start:int=None, worker_port_end:int=None) -> tuple[str,str]:
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostname, username=username, password=password)

    worker_port_arg = ""
    if worker_port_start is not None:
        if worker_port_end is not None:
            worker_port_arg = f"--worker-port {worker_port_start}:{worker_port_end}"
        else:
            worker_port_arg = f"--worker-port {worker_port_start}"

    command = (
        "docker exec finance_scheduler bash -c "
        "'cd /opt/dagster/app && "
        "source .venv/bin/activate && "
        f"nohup dask worker tcp://{scheduler_ip}:{scheduler_port} --nthreads 1 "
        f"--nworkers {nworkers} --memory-limit {mem_limit} "
        f"--host 0.0.0.0 {worker_port_arg} "
        f"> /tmp/dask_worker.log 2>&1 &'"
    )
    stdin, stdout, stderr = ssh.exec_command(command)
    dagster_logger.info(f"stdin:{stdin} | stdout:{stdout} | stderr:{stderr}")
    dagster_logger.info(f"Dask Worker created on scheduler machine : {hostname} / {nworkers} workers with {mem_limit} mem. each")
    time.sleep(1)
    # Fetch the worker log
    _, log_out, log_err = ssh.exec_command("tail -n 75 /tmp/dask_worker.log")
    print(log_out.read().decode(), log_err.read().decode())
    return log_out, log_err

def create_dask_worker(hostname: str, username: str, password: str,
                       scheduler_ip: str = 'localhost', scheduler_port: int = 8786,
                       nworkers: int = 1, mem_limit: str = '1GB',
                       worker_port_start: int = None, worker_port_end: int = None,
                       nanny_port_start: int = None, nanny_port_end: int = None) -> tuple[str,str]:
    """
    :param scheduler_port: Port the scheduler is listening on (default 8786)
    :param worker_port_start: Starting port for worker processes (required if nworkers > 1)
    :param worker_port_end: Ending port for worker processes (required if nworkers > 1)
    :param nanny_port_start: Starting port for nanny processes (optional)
    :param nanny_port_end: Ending port for nanny processes (optional)
    """
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostname, username=username, password=password)

    # Build worker port argument
    worker_port_arg = ""
    if worker_port_start is not None:
        if worker_port_end is not None:
            worker_port_arg = f"--worker-port {worker_port_start}:{worker_port_end}"
        else:
            worker_port_arg = f"--worker-port {worker_port_start}"

    # Build nanny port argument
    nanny_port_arg = ""
    if nanny_port_start is not None:
        if nanny_port_end is not None:
            nanny_port_arg = f"--nanny-port {nanny_port_start}:{nanny_port_end}"
        else:
            nanny_port_arg = f"--nanny-port {nanny_port_start}"

    command = (
        "cd /home/jay/PycharmProjects/finance/ && "
        "set -a && source questrade.env && source .env && set +a && "
        ". .venv/bin/activate && "
        f"nohup dask worker tcp://{scheduler_ip}:{scheduler_port} "
        f"--nthreads 1 --nworkers {nworkers} --memory-limit {mem_limit} "
        f"--host 0.0.0.0 {worker_port_arg} {nanny_port_arg} "
        f"> /tmp/dask_worker.log 2>&1 < /dev/null &"
    )

    # Start the worker in background
    ssh.exec_command(command)
    # Give it a moment to start
    time.sleep(2)
    # Fetch the last few lines of the worker log
    stdin, stdout, stderr = ssh.exec_command("tail -n 75 /tmp/dask_worker.log")
    log_out = stdout.read().decode()
    log_err = stderr.read().decode()
    dagster_logger.info(f"Dask worker log:\n{log_out}")
    dagster_logger.error(f"Dask worker errors (if any):\n{log_err}")
    return log_out, log_err

def kill_and_redeploy_dask_home_setup() -> None:
    kill_dask_deployment_home_setup()
    time.sleep(10)
    deploy_dask_home_setup()


def deploy_dask_home_setup() -> None:
    scheduler_port = 8786
    scheduler_ip = os.getenv('cluster_server_1_address') or "localhost"
    user0 = os.getenv('cluster_server_1_username') or "admin"
    password0 = os.getenv('cluster_server_1_password') or "password"
    create_dask_scheduler(
        hostname=scheduler_ip,
        username=user0,
        password=password0,
        port=scheduler_port
    )

    create_dask_worker_on_scheduler(hostname=scheduler_ip, username=user0,
                                    password=password0, scheduler_ip="localhost",
                                    nworkers=30, mem_limit='9GB',
                                    worker_port_start=9000,  # Specify port range for 30 workers
                                    worker_port_end=9029)
    # # create worker 2
    # create_dask_worker(
    #     hostname=os.getenv('cluster_server_2_address'),
    #     username=os.getenv('cluster_server_2_username'),
    #     password=os.getenv('cluster_server_2_password'),
    #     scheduler_ip=scheduler_ip,  # Use actual IP, not localhost
    #     scheduler_port=scheduler_port,
    #     nworkers=7,
    #     mem_limit='6GB',
    #     worker_port_start=9000,  # Port range for 7 workers
    #     worker_port_end=9006
    # )
    # create worker 3, on work PC
    address = os.getenv("cluster_server_0_address") or "localhost"
    user = os.getenv("cluster_server_0_username") or "admin"
    password = os.getenv("cluster_server_0_password") or "password"
    create_dask_worker(
        hostname=address,
        username=user,
        password=password,
        scheduler_ip=scheduler_ip,  # Use actual IP, not localhost
        scheduler_port=scheduler_port,
        nworkers=2,
        mem_limit='10GB',
        worker_port_start=9000,  # Port range for 5 workers
        worker_port_end=9001,
    )

    time.sleep(9)

def kill_dask_deployment_home_setup():
    # server 0 is used for task scheduler and the rest used for dask
    for server_number in [0,1,2]:
        dagster_logger.info('Trying to kill dask processes @: '+os.getenv(f'cluster_server_{server_number}_address'))
        pids_with_dask_before = get_remote_process_ids_ubuntu(os.getenv(f'cluster_server_{server_number}_address'), os.getenv(f'cluster_server_{server_number}_username'), os.getenv(f'cluster_server_{server_number}_password'), 'dask')
        dagster_logger.info(pids_with_dask_before)
        kill_remote_ubuntu_process_ids(os.getenv(f'cluster_server_{server_number}_address'), os.getenv(f'cluster_server_{server_number}_username'), os.getenv(f'cluster_server_{server_number}_password'), pids_with_dask_before)
        time.sleep(5)
        pids_with_dask_after =  get_remote_process_ids_ubuntu(os.getenv(f'cluster_server_{server_number}_address'), os.getenv(f'cluster_server_{server_number}_username'), os.getenv(f'cluster_server_{server_number}_password'), 'dask')
        ctr = 0
        while len(pids_with_dask_after) > 1:
            dagster_logger.info(f"There are still dask processes / {pids_with_dask_after}")
            time.sleep(5)
            pids_with_dask_after =  get_remote_process_ids_ubuntu(os.getenv(f'cluster_server_{server_number}_address'), os.getenv(f'cluster_server_{server_number}_username'), os.getenv(f'cluster_server_{server_number}_password'), 'dask')
            ctr = ctr + 1
            if ctr == 3:
                kill_remote_ubuntu_process_ids(os.getenv(f'cluster_server_{server_number}_address'),
                                               os.getenv(f'cluster_server_{server_number}_username'),
                                               os.getenv(f'cluster_server_{server_number}_password'),
                                               pids_with_dask_after)
            if ctr == 6:
                raise RuntimeError('Could not kill remote processes')

        dagster_logger.info(pids_with_dask_after)
        # for each machine, ensure that there are fewer processes and that only the two grep/ps processes are running
        assert len(pids_with_dask_before) >= len(pids_with_dask_after), "There are more dask tasks now than before"
        assert len(pids_with_dask_after) < 3, "There are more than grep | ps process running dask"

def upload_files_to_dask_cluster():
    with sqlalchemy.create_engine(os.getenv("home_connection_string")).begin() as conn:
        running_cluster_location = pd.read_sql("""SELECT var, value FROM
         environment_variables WHERE
         var = 'distributed_dask_cluster'""",  conn)['value'].values[0]
        client = dask.distributed.get_client(running_cluster_location)
        print(f"Using Dask Cluster: {str(client)}")
        client.upload_file('/home/jay/PycharmProjects/finance/finance_data_eng.py')
        print(client)

def process_list_with_dask(input_list: list[Any], func: Callable, num_splits: int, cluster:CLUSTER_TYPE = 'DISTRIBUTED', priority:int = 1, kwargs: Optional[dict[str, Any]] = None,) -> None:
    """
    Split the input list into multiple sublists using Dask and execute the provided function on each split list.

    :param input_list: The input list to be split.
    :param func: The function to be executed on each split list.
    :param num_splits:  The number of splits to create from the input list.
    :param cluster:  Either "local" or "distributed". This function chooses the cluster to send the tasks to.
    :param kwargs:  Sometimes the func needs extra kwargs, this should be a dict if present, else None
    :param priority:  priority, an int with default 1. Higher has a higher priority and will run first on the cluster.
    :return: None
    """

    # Calculate the size of each split
    split_size = len(input_list) // num_splits
    split_size = max(1, split_size)

    # Create the splits using Dask
    splits = [input_list[i:i + split_size] for i in range(0, len(input_list), split_size)]
    dagster_logger.info(f"{num_splits} splits each of around {split_size} size made. {len(input_list)} in total.")
    if cluster == 'DISTRIBUTED':
        running_cluster_location = pd.read_sql("""SELECT var, value FROM
         environment_variables WHERE
         var = 'distributed_dask_cluster'""",  sqlalchemy.create_engine(os.getenv('home_connection_string')))['value'].values[0]
        client = dask.distributed.get_client(running_cluster_location)
        dagster_logger.info(str(running_cluster_location))
        dagster_logger.info(f"Using Distributed Dask Cluster : {str(client)}")

    elif cluster == 'LOCAL':

        try:
            dagster_logger.info("Checking for Running Local Dask Cluster")
            client = dask.distributed.get_client(f"tcp://{os.getenv('local_db_address')}:8786")
            dagster_logger.info(f"Client acquired at tcp://{os.getenv('local_db_address')}:8786")
            client.shutdown()
            dagster_logger.info("Client shutdown")
            time.sleep(15)
        except Exception:
            dagster_logger.info(error_handling())
            dagster_logger.info("No Client found")

        dagster_logger.info('Creating Local Dask Cluster')
        abc = execute_script_with_cmd("/home/jay/PycharmProjects/finance/src/finance/common/finance_launch_dask_cluster.py")
        dagster_logger.info(str(abc))

        running_cluster_location = pd.read_sql("""SELECT var, value FROM
         environment_variables WHERE
         var = 'local_dask_cluster'""",  sqlalchemy.create_engine(os.getenv('home_connection_string')))['value'].values[0]
        client = dask.distributed.get_client(running_cluster_location)
        dagster_logger.info(str(running_cluster_location))
        dagster_logger.info(f"Using Local Dask Cluster : {str(client)}")

    # Create Dask delayed objects for each split and apply the provided function
    if kwargs:
        delayed_results = [dask.delayed(func)(split, **kwargs) for split in splits]
    else:
        delayed_results = [dask.delayed(func)(split) for split in splits]

    with sqlalchemy.create_engine(os.getenv('home_connection_string')).begin() as conn:
        running_cluster_location = pd.read_sql("""SELECT var, value FROM
         environment_variables WHERE
         var = 'distributed_dask_cluster'""",  conn)['value'].values[0]
        client = dask.distributed.get_client(running_cluster_location)

    # Compute the results using Dask's parallel processing capabilities
    output = dask.compute(*delayed_results, priority=priority, )

    return output


def find_number_of_free_dask_workers():
    with sqlalchemy.create_engine(os.getenv('home_connection_string')).begin() as conn:
        running_cluster_location = pd.read_sql(""" SELECT var, value
                                                   FROM environment_variables
                                                   WHERE var = 'distributed_dask_cluster' """, conn)['value'].values[0]

        # If running inside Docker container, use localhost instead of host IP
        if os.path.exists('/.dockerenv'):  # Check if inside Docker
            # Replace host IP with localhost
            if '192.168.100.54' in running_cluster_location:
                running_cluster_location = running_cluster_location.replace('192.168.100.54', 'localhost')

        client = get_client(running_cluster_location)
        pprint(f"Using Dask Cluster: {str(client)}")
        abc = client.processing()
        done = [1 for v in abc.values() if len(v) == 0]
        return sum(done)



import datetime
import itertools
import json
import logging
import mimetypes
import os
import re
import smtplib
import subprocess
import threading
import time
import traceback
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pprint import pprint
from typing import Any, Iterable, Optional

import bs4 as bs
import numpy as np
import pandas as pd
import requests  # type: ignore
import selenium
import tornado.autoreload
import tornado.ioloop
from selenium.webdriver.chrome.options import Options as chrome_options
from selenium.webdriver.firefox.options import Options as firefox_options

path = os.path.abspath(os.path.dirname(__file__)).replace("\\", "/")+ "/"
data_path : str = path + 'data/'
today = datetime.datetime.today()

# logger
dagster_logger = logging.getLogger("dagster_logger")
dagster_logger.setLevel(logging.INFO)
# create console handler
handler = logging.StreamHandler()
# create formatter and add it to the handler
formatter = logging.Formatter('%(asctime)s, %(message)s')
handler.setFormatter(formatter)
# add the handler to the logger
dagster_logger.addHandler(handler)
# create console handler
handler2 = logging.FileHandler(path + "dagster_logger.log")
# add the handler to the logger
dagster_logger.addHandler(handler2)

# BOKEH TOOLS
BOKEH_TOOLS = "pan,wheel_zoom,box_zoom,reset,save"

# Utilities
def set_env(secrets_:dict[str, str]) -> None:
    for x,y in secrets_.items():
        os.environ[str(x)] = str(y)

def convert_large_num_to_human_read(num: int) -> str:
    if num // 10**12 >= 1:
        return f"{(num // 10**12) + (num%10**12)/10**12:,.1f} T"
    elif num // 10**9 >= 1:
        return f"{(num // 10**9) + (num%10**9)/10**9:,.1f} B"
    elif num // 10**6 >= 1:
        return f"{(num // 10**6) + (num%10**6)/10**6:,.1f} M"
    else:
        return f"{int(num):,}"

def write_to_env_file(env_dict: dict[str, Any], file_path, prefix=None):
    """
    Writes the dictionary as environment variables with 'xyz_' prefix into a .env file.

    Parameters:
    - env_dict (dict): Dictionary containing the environment variables.
    - file_path (str): Path where the .env file should be created.

    Example:
    - env_dict = {
        'access_token': 'ccc',
        'api_server': 'https://abc.com/',
        'expires_in': 1800,
        'refresh_token': 'abc',
        'token_type': 'xxx'
    }
    """
    with open(file_path, 'w') as file:
        for key, value in env_dict.items():
            # Add 'xyz_' prefix to the variable names
            env_var_name = key
            if prefix:
                env_var_name = f"{prefix}{key}"
            # Write the variable to the .env file in the format VAR_NAME=VALUE
            file.write(f"{env_var_name}={value}\n")

    print(f".env file created at {file_path}")

def read_json_from_api(url:str, headers:Optional[dict]=None, data:Optional[dict[str, Any]]=None, session: requests.Session = None) -> tuple[Optional[dict], dict[str, Any]]:
    output_data = None
    rate_limit_dict = {"remaining_requests": None, "reset_timestamp_utc": None}
    try:
        if headers:
            if session:
                if data:
                    response = session.get(url, headers=headers, json=data)
                else:
                    response = session.get(url, headers=headers)
            else:
                if data:
                    response = requests.get(url, headers=headers, json=data)
                else:
                    response = requests.get(url, headers=headers)
        else:
            if session:
                if data:
                    response = session.get(url, json=data)
                else:
                    response = session.get(url)

            else:
                if data:
                    response = requests.get(url, json=data)
                else:
                    response = requests.get(url)
        output_data = response.json()
        response.raise_for_status()  # Raise an exception for 4xx or 5xx status codes
        if response.headers:
            if response.headers._store:
                if 'x-ratelimit-remaining' in response.headers._store.keys():
                    if response.headers._store['x-ratelimit-remaining']:
                        _, val =  response.headers._store['x-ratelimit-remaining']
                        rate_limit_dict['remaining_requests'] = val
                    if response.headers._store['x-ratelimit-reset']:
                        _, val = response.headers._store['x-ratelimit-reset']
                        rate_limit_dict['reset_timestamp_utc'] = val
    except requests.exceptions.RequestException as e:
        print("Error message:", e)
        if output_data:
            print("Failed to make API call.")
            pprint(output_data)
        output_data = {'error': str(e)}
    return output_data, rate_limit_dict

def read_json_from_api_post(url:str, headers:Optional[dict]=None, data:Optional[dict[str, Any]]=None, session: requests.Session = None) -> tuple[Optional[dict], dict[str, Any]]:
    output_data = None
    rate_limit_dict = {"remaining_requests": None, "reset_timestamp_utc": None}
    try:
        if headers:
            if session:
                if data:
                    response = session.post(url, headers=headers, json=data)
                else:
                    response = session.post(url, headers=headers)
            else:
                if data:
                    response = requests.post(url, headers=headers, json=data)
                else:
                    response = requests.post(url, headers=headers)
        else:
            if session:
                if data:
                    response = session.post(url, json=data)
                else:
                    response = session.post(url)

            else:
                if data:
                    response = requests.post(url, json=data)
                else:
                    response = requests.post(url)
        output_data = response.json()
        response.raise_for_status()  # Raise an exception for 4xx or 5xx status codes
        if response.headers:
            if response.headers._store:
                if response.headers._store['x-ratelimit-remaining']:
                    _, val =  response.headers._store['x-ratelimit-remaining']
                    rate_limit_dict['remaining_requests'] = val
                if response.headers._store['x-ratelimit-reset']:
                    _, val = response.headers._store['x-ratelimit-reset']
                    rate_limit_dict['reset_timestamp_utc'] = val
    except requests.exceptions.RequestException as e:
        print("Error message:", e)
        if output_data:
            print("Failed to make API call.")
            pprint(output_data)
    return output_data, rate_limit_dict

def above_zero_numpy_mask(series):
    # Convert the Pandas Series to a NumPy array for vectorized operations
    array = np.array(series)
    # Create a mask for values greater than 0
    mask = array > 0
    # Apply the vectorization condition to get 1 for values greater than 0 and 0 otherwise
    array[mask] = 1
    array[~mask] = 0
    # Convert the array back to a Pandas Series
    vectorized_series = pd.Series(array.astype(int), index=series.index)
    return vectorized_series

def flatten_list(array: pd.DataFrame | list) -> list:
    if isinstance(array[0], list):
        return [item for sublist in array for item in sublist]
    elif isinstance(array[0], pd.DataFrame):
        return array
    else:
        return array

def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    new_iter = list(iterable)  # allows duplicate elements
    return itertools.chain.from_iterable(itertools.combinations(new_iter, r) for r in range(1, len(new_iter) + 1))

def do_work(list_of_tasks:Iterable, wait_time:float=1) -> int:
    """
    This function simulates doing work and returning a number
    :param list_of_tasks: Iterable for a list of tasks to do that take *wait_time* seconds
    :param wait_time: Float or Int for the amount of time the task should take in seconds
    :return: a randomly generated integer between 1 and 10
    """
    for x in list_of_tasks:
        dagster_logger.info(str(x))
        time.sleep(wait_time + np.random.randint(0,10))

    return np.random.randint(0,10)

def find_parent_of_src_folder(start_dir:Optional[str]) -> Optional[str]:
    # Walk upwards through the directory structure
    if start_dir:
        current_path = start_dir
    else:
        current_path = os.getcwd()
    while current_path != os.path.dirname(current_path):  # stop if we reach the root
        if os.path.basename(current_path) == 'src':
            # Return the parent of the 'src' folder
            return os.path.dirname(current_path)
        current_path = os.path.dirname(current_path)  # go up one directory level
    return None  # return None if no 'src' folder is found

def convert_unix_timestamp_to_pandas_date(input_date):
    return pd.to_datetime(datetime.datetime.fromtimestamp(input_date/1000).date())

def parse_date_features(idf_: pd.DataFrame, holidays_: Optional[list] = None) -> pd.DataFrame:
    """
    This function adds several datetime-like features to the input dataframe and returns the dataframe. Optionally, a
    column is made if holiday dates are passed in. This function also creates cyclical hour and month variables.
    :param idf_: an input dataframe for ML that has a datetime index
    :param holidays_: a list of holidays for that region, this parameter is optional
    :return: the dataframe with new datetime-specific columns
    """
    idf = idf_.copy()
    idf.index = pd.to_datetime(idf.index)
    idf["hour"] = idf.index.hour
    idf["month"] = idf.index.month
    idf["year"] = idf.index.year
    idf["weekday"] = idf.index.weekday  # 0 is sunday
    idf["hour_sin"] = np.sin(idf.hour * (2.0 * np.pi / 24))
    idf["hour_cos"] = np.cos(idf.hour * (2.0 * np.pi / 24))
    idf["month_sin"] = np.sin((idf.month - 1) * (2.0 * np.pi / 12))
    idf["month_cos"] = np.cos((idf.month - 1) * (2.0 * np.pi / 12))
    if holidays_:
        idf["holiday"] = [1 if x in holidays_ else 0 for x in idf.index]
    return idf

def wrap_in_paragraphs(txt:str, colour:str="DarkSlateBlue", size:int=4) -> str:
    """
    This function wraps text in paragraph, bold and font tags - according to the colour and size given.
    :param text: text to wrap in tags_
    :param colour: colour of the font
    :param size: size of the font
    :return: string wrapped in html tags
    """
    return f"""<p><b><font color={colour} size={size}>{txt}</font></b></p>"""

def sendemail_(TEXT:str, HTML:str, SUBJECT:str = 'Daily Market Update', filename: Optional[str] = None, TO: Iterable = (os.getenv('my_email'),)):
    """

    This function sends emails to the email list depending on the para

    :param TEXT: text to send in an email
    :param HTML: text to send in an email, but in HTML (default)
    :param SUBJECT: integer that indicates if this is the email sent weekly (start of the week, monday at 7am)
    :param filename: if a file name is present, attach the file to the email and send it
    :return:
    """
    # This is a temporary fix. Be careful of malicious links
    # context = ssl._create_unverified_context()
    TO = list(TO)

    senders_email = os.getenv('senders_email', 'email')  # senders email
    senders_pswd = os.getenv('senders_pw', 'pw')# senders password

    # current date, and a date 5 days away
    curtime = datetime.datetime.now().date()
    # Gmail Sign In
    gmail_sender = senders_email
    gmail_passwd = senders_pswd

    msg = MIMEMultipart('alternative')  # tell the package we'd prefer HTML emails
    msg['Subject'] = SUBJECT  # set the SUBJECT of the email
    msg['From'] = gmail_sender  # set the FROM field of the email
    msg['To'] = ', '.join(TO)  # set the TO field of the email

    # add the 2 parts of the email (one plain text, one html)
    part1 = MIMEText(TEXT, 'plain')
    part2 = MIMEText(HTML, 'html')
    # It will default to the plain text verison if the HTML doesn't work, plain must go first
    msg.attach(part1)
    msg.attach(part2)

    if filename:
        ctype, encoding = mimetypes.guess_type(filename)
        if ctype is None or encoding is not None:
            ctype = "application/octet-stream"
        maintype, subtype = ctype.split("/", 1)

        with open(filename) as fp:
            attachment = MIMEText(fp.read(), _subtype=subtype)
        attachment.add_header("Content-Disposition", "attachment", filename=os.path.basename(filename))
        msg.attach(attachment)

    # connect to the GMAIL server
    server = smtplib.SMTP_SSL('smtp.gmail.com', 465)
    # login to the GMAIL server
    server.login(gmail_sender, gmail_passwd)

    try:
        # send email and confirm email is sent / time it is sent
        server.sendmail(gmail_sender, TO, msg.as_string())
        logging.info(str(curtime) + ' email sent')
    except Exception as e:
        # print error if not sent, and confirm it wasn't sent
        dagster_logger.info(str(e))
        dagster_logger.info(error_handling())
        dagster_logger.info(str(curtime) + ' error sending mail')

    server.quit()

def update_json_file(file_location: str, dictionary: dict) -> None:
    """Updates a JSON file with a dictionary key.

    Args:
        file_location (str): The file path of the JSON file.
        dictionary (dict): The dictionary to update the JSON file with.

    Returns:
        None: The function does not return a value.

    Raises:
        FileNotFoundError: If the JSON file is not found.
        json.JSONDecodeError: If the JSON file has an invalid format.
        Exception: If any other error occurs during the update process.
    """
    try:
        # Load the JSON file
        with open(file_location, 'r') as file:
            data = json.load(file)

        # Update the JSON data with the dictionary
        data.update(dictionary)

        # Write the updated data back to the JSON file
        with open(file_location, 'w') as file:
            json.dump(data, file, indent=4)

        dagster_logger.info("JSON file updated successfully.")
    except FileNotFoundError:
        dagster_logger.info("File not found.")
    except json.JSONDecodeError:
        dagster_logger.info("Invalid JSON file format.")
    except Exception as e:
        dagster_logger.info(f"An error occurred: {str(e)}")

def error_handling() -> str:
    """
    This function returns a string with all of the information regarding the error
    :return: string with the error information
    """
    return traceback.format_exc()

# Selenium / Chrome-related
def get_chromedriver_download_table():
    url = "https://googlechromelabs.github.io/chrome-for-testing/known-good-versions-with-downloads.json"
    response = read_json_from_api(url)
    ok = []
    for ver in response['versions']:
        try:
            for w in ver['downloads']['chromedriver']:
                ok.append({"revision": ver['revision'], 'version': ver['version'], 'platform': w['platform'], 'url':w['url']})
        except KeyError:
            pass
    wow = pd.DataFrame(ok)
    wow['timestamp'] = response['timestamp']
    return wow

def check_selenium_browser_driver_ubuntu(path_to_chromedriver_dir):
    version = get_application_version_ubuntu()
    vers = version.split('.')[0]
    assert f"chromedriver{vers}" in os.listdir(path_to_chromedriver_dir + "chromedriver/")[0], f"in_folder= {os.listdir(path_to_chromedriver_dir + "chromedriver/")}, path= {path_to_chromedriver_dir}, version= {vers}, chromedriver{vers}.exe not in chromedriver folder"
    dagster_logger.info(f"ChromeDriver version and Browser version are synced {vers}")

def get_application_version_ubuntu(path_to_exe=r"google-chrome"):
    command = rf"""{path_to_exe} --version"""
    version = ''.join([b for b in os.popen(command).readlines()])
    version = re.findall(r"\s(\d+.+)", version, )[0].strip()
    return version

def get_selenium_driver(browser="chrome", dat_path='/home/jay/PycharmProjects/home/data/'):
    if browser == 'chrome':
        vers = get_application_version_ubuntu().split('.')[0]
        print(vers)
        chromeOptions = chrome_options()
        chromeOptions.binary_location = '/usr/bin/google-chrome'
        chromeOptions.add_experimental_option("prefs", {
            "download.default_directory": r"/home/jay/Downloads",
            "download.prompt_for_download": False,
            "download.directory_upgrade": True,
            "safebrowsing.enabled": True
        })
        chromeOptions.add_argument("--headless")
        chromeOptions.add_argument('--no-sandbox')
        chromeOptions.add_argument('--disable-dev-shm-usage')
        # chrome_driver_path = dat_path + 'chromedriver/chromedriver' + vers
        driver = selenium.webdriver.Chrome(options=chromeOptions)
    else:
        firefoxOptions = firefox_options()
        firefoxOptions.set_preference("browser.download.folderList", 2)
        firefoxOptions.set_preference("browser.download.manager.showWhenStarting", False)
        firefoxOptions.set_preference("browser.download.dir", path.replace('/', '\\') + 'data\\downloads\\')
        firefoxOptions.set_preference("browser.helperApps.neverAsk.saveToDisk",
                                      "application/octet-stream,application/vnd.ms-excel,application/vnd.openxmlformats-officedocument.spreadsheetml.sheet")
        driver = selenium.webdriver.Firefox(options=firefoxOptions)
    return driver

def grab_soup(url_, driver):
    """
    This function enables a driver (using Firefox or Chrome), goes to the URL, and retrieves the data after the JS is loaded.

    :param url_: url to go to to retrieve data
    :param browser: browser to use, defaults to firefox (requires geckodriver.exe on path)
    :return:

    soup - the data of the page
    driver - the browser (process) instance
    """
    driver.get(url_)  # go to the URL
    _ = driver.page_source
    time.sleep(1)  # sleep for 1 second  to ensure all JS scripts are loaded
    html = driver.execute_script("return document.body.outerHTML;")  # execute javascript code
    soup_ = bs.BeautifulSoup(html, 'lxml')  # read the data as html (using lxml driver)
    return soup_

def xpath_soup(element):
    """
    Generate xpath of soup element
    :param element: bs4 text or node
    :return: xpath as string
    """
    components = []
    child = element if element.name else element.parent
    for parent in child.parents:
        """
        @type parent: bs4.element.Tag
        """
        previous = itertools.islice(parent.children, 0, parent.contents.index(child))
        xpath_tag = child.name
        xpath_index = sum(1 for i in previous if i.name == xpath_tag) + 1
        components.append(xpath_tag if xpath_index == 1 else '%s[%d]' % (xpath_tag, xpath_index))
        child = parent
    components.reverse()
    return '/%s' % '/'.join(components)


# BOKEH
def run_bokeh_app(files, port=5002, new='tab'):
    # all files in directory that you want as bokeh apps
    # run_bokeh_app(["finance_charts.py",])
    # run_bokeh_app(["personal.py", ])

    def create_bokeh_server(io_loop, files, argvs, host, port):
        '''Start bokeh server with applications paths'''
        from bokeh.command.util import build_single_handler_applications
        from bokeh.server.server import Server

        # Turn file paths into bokeh apps
        apps = build_single_handler_applications(files, argvs)

        # kwargs lifted from bokeh serve call to Server, with created io_loop
        kwargs = {
            'io_loop': io_loop,
            'generate_session_ids': True,
            'redirect_root': True,
            'use_x_headers': False,
            'secret_key': None,
            'num_procs': 1,
            'host': host,
            'sign_sessions': False,
            'develop': False,
            'port': port,
            'use_index': True
        }
        server = Server(apps, **kwargs)

        return server

    def start_bokeh(io_loop):
        '''Start the `io_loop`'''
        io_loop.start()
        return None

    def launch_app(host, app_name, new):
        '''Lauch app in browser

        Ideally this would `bokeh.util.browser.view()`, but it doesn't work
        '''
        import webbrowser

        # Map method strings to webbrowser method
        options = {'current':0, 'window':1, 'tab':2}

        # Concatenate url and open in browser, creating a session
        app_url = 'http://{}/{}'.format(host, app_name)
        print('Opening `{}` in browser'.format(app_url))
        webbrowser.open(app_url, new=options[new])

        return None

    def server_loop(server, io_loop):
        '''Check connections once session created and close on disconnect'''
        import time

        connected = [True,]
        session_loaded = False
        while any(connected):

            # Check if no session started on server
            sessions = server.get_sessions()
            if not session_loaded:
                if sessions:
                    session_loaded = True
            # Once 1+ sessions started, check for no connections
            else:
                # List of bools for each session
                connected = [True,]*len(sessions)
                # Set `connected` item false no connections on session
                for i in range(len(sessions)):
                    if sessions[i].connection_count == 0:
                        connected[i] = False
            # Keep the pace down
            time.sleep(2)

        # Stop server once opened session connections closed
        io_loop.stop()

        return None



    # Initialize some values, sanatize the paths to the bokeh plots
    argvs = {}
    app_names = []
    for path in files:
        argvs[path] = None
        app_names.append(os.path.splitext(os.path.split(path)[1])[0])

    # Concate hostname/port for creating handlers, launching apps
    host = 'localhost:{}'.format(port)
    # Initialize the tornado server
    io_loop = tornado.ioloop.IOLoop.instance()
    # tornado.autoreload.start(io_loop)

    # Add the io_loop to the bokeh server
    server = create_bokeh_server(io_loop, files, argvs, host, port)

    print('Starting the server on {}'.format(host))
    args = (io_loop,)
    th_startup = threading.Thread(target=start_bokeh, args=args)
    th_startup.start()

    # Launch each application in own tab or window
    th_launch = [None,]*len(app_names)
    for i in range(len(app_names)):
        args = (host, app_names[i], new)
        th_launch[i] = threading.Thread(target=launch_app, args=args)
        th_launch[i].start()
        # Delay to allow tabs to open in same browser window
        time.sleep(2)

    # Run session connection test, then stop `io_loop`
    args = (server, io_loop)
    th_shutdown = threading.Thread(target=server_loop, args=args)
    th_shutdown.start()

# DATE HELPERS
def split_date_range_into_months(start_date: pd.Timestamp, end_date: pd.Timestamp) -> list[tuple[str,str]]:
    """Splits the range from start_date to end_date into monthly ranges."""
    months = []
    current_start = start_date.replace(day=1)

    while current_start <= end_date:
        # Get the last day of the current month
        current_end = (current_start + pd.offsets.MonthEnd(0)).date()  # Last day of the current month

        # Ensure we do not exceed the overall end_date
        current_end = min(current_end, end_date.date())

        # Append the current month's range as a tuple
        months.append((current_start.strftime('%Y-%m-%d'), current_end.strftime('%Y-%m-%d')))

        # Move to the next month
        current_start = (current_start + pd.DateOffset(months=1)).replace(day=1)

    return months

def convert_column_name(column_name: str) -> str:
    """
    Converts a camelCase or PascalCase string into snake_case.

    Args:
    - column_name (str): The input column name in camelCase or PascalCase.

    Returns:
    - str: The column name converted to snake_case.
    """
    # Use regular expression to add underscores before capital letters and convert them to lowercase
    result = re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', column_name).lower()
    result = result.replace(" ", "_")
    return result.strip()

def is_business_day(date: datetime.date) -> bool:
    return bool(len(pd.bdate_range(date, date)))

def date_conversion(date: str | datetime.datetime | None, return_type: str = 'str'):
    if date is None:
        return None
    elif isinstance(date, str):
        try:
            date_obj = datetime.datetime.fromisoformat(date)
        except ValueError:
            raise ValueError(f"Invalid date string format: {date}. Expected ISO format (YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS).")
    elif isinstance(date, datetime.datetime):
        date_obj = date
    else:
        raise TypeError(f"Unsupported date type: {type(date)}")

    if return_type == "str":
        return date_obj.isoformat()
    elif return_type == "short":
        return date_obj.strftime("%Y-%m-%d")
    elif return_type == "obj":
        return date_obj
    elif return_type == "timestamp":
        return pd.to_datetime(date_obj)
    else:
        raise ValueError(f"Invalid return_type: {return_type}. Expected 'str' or 'datetime'.")


# WINDOWS FUNCS
def get_application_version_windows(
        path_to_exe=r"C:\\Program Files\\BraveSoftware\\Brave-Browser\\Application\\brave.exe"):
    command = rf""" wmic datafile where name="{path_to_exe}" get Version /value """
    version = ''.join([b for b in os.popen(command).readlines()])
    version = version.strip().split('=')[1]
    return version

def execute_cmd_windows(commands: str) -> Optional[str]:
    try:
        out_var = subprocess.Popen(commands, shell=True)
        dagster_logger.info(f"{commands} | Running...")

        (output, err) = out_var.communicate()
        output_str = None
        err_str = None
        # Decode the output and error from bytes to string
        # Print the output and error
        if output:
            output_str = output.decode('utf-8')
            print("Output:")
            # This will give you the output of the command being executed
            dagster_logger.info("Command output: " + output_str)

        if err:
            err_str = err.decode('utf-8')
            print("Error:")
            print(err_str)

        return output_str
    except Exception:
        dagster_logger.info(error_handling())
        return "Did not run"

def check_selenium_browser_driver_windows():
    version = get_application_version_windows()
    assert f"chromedriver{version.split('.')[0]}.exe" in os.listdir(data_path + "chromedriver/"), f"version = {version}, chromedriver{version.split('.')[0]}.exe not in chromedriver folder"
    dagster_logger.info(f"ChromeDriver version and Browser version are synced {version}")



from .jragbeer_common_dask import CLUSTER_TYPE as CLUSTER_TYPE
from .jragbeer_common_dask import \
    create_dask_scheduler as create_dask_scheduler
from .jragbeer_common_dask import create_dask_worker as create_dask_worker
from .jragbeer_common_dask import \
    deploy_dask_home_setup as deploy_dask_home_setup
from .jragbeer_common_dask import \
    find_number_of_free_dask_workers as find_number_of_free_dask_workers
from .jragbeer_common_dask import \
    kill_and_redeploy_dask_home_setup as kill_and_redeploy_dask_home_setup
from .jragbeer_common_dask import \
    kill_dask_deployment_home_setup as kill_dask_deployment_home_setup
from .jragbeer_common_dask import \
    process_list_with_dask as process_list_with_dask
from .jragbeer_common_dask import \
    update_dask_environment_vars_local as update_dask_environment_vars_local
from .jragbeer_common_dask import \
    upload_files_to_dask_cluster as upload_files_to_dask_cluster
